# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import modules, os.path

from gui        import fileChooser, errorMsgBox, infoMsgBox
from tools      import consts, prefs
from media      import track
from gettext    import gettext as _
from os.path    import isdir, basename, dirname
from gui.window import Window


# Module configuration
MOD_NAME            = _('Status File')
MOD_DESC            = _('Generate a simple text file with the current status')
MOD_IS_MANDATORY    = False
MOD_IS_CONFIGURABLE = True


# Default preferences
PREFS_DEFAULT_FILE   = os.path.join(consts.dirCfg, 'now-playing.txt')
PREFS_DEFAULT_STATUS = 'Now playing {title}\nby {artist}'


class StatusFile(modules.ThreadedModule):
    """ Allow external program to display the current status """


    def __init__(self, wTree):
        """ Constructor """
        modules.ThreadedModule.__init__(self)
        self.window = None
        modules.register(self, [modules.MSG_EVT_NEW_TRACK, modules.MSG_EVT_STOPPED, modules.MSG_EVT_MOD_UNLOADED, modules.MSG_EVT_APP_QUIT])


    def configure(self, parent):
        """ Show the configuration window """
        if self.window is None:
            self.window    = Window('StatusFile.glade', 'vbox1', __name__, MOD_NAME, 355, 345)
            self.btnOk     = self.window.getWidget('btn-ok')
            self.txtFile   = self.window.getWidget('txt-file')
            self.txtStatus = self.window.getWidget('txt-status')
            # GTK handlers
            self.btnOk.connect('clicked', self.onOk)
            self.window.getWidget('btn-help').connect('clicked', self.onHelp)
            self.window.getWidget('btn-cancel').connect('clicked', lambda btn: self.window.hide())
            self.window.getWidget('btn-open').connect('clicked', self.onOpen)

        # Fill current preferences
        if not self.window.isVisible():
            self.txtFile.set_text(prefs.get(__name__, 'file', PREFS_DEFAULT_FILE))
            self.txtStatus.get_buffer().set_text(prefs.get(__name__, 'status', PREFS_DEFAULT_STATUS))
            self.btnOk.grab_focus()

        self.window.show()


    def updateFile(self, track):
        """ Show the notification based on the given track """
        output = open(prefs.get(__name__, 'file', PREFS_DEFAULT_FILE), 'w')

        if track is None: output.write('')
        else:             output.write(track.format(prefs.get(__name__, 'status', PREFS_DEFAULT_STATUS)))

        output.close()


    # --== GTK handlers ==--


    def onHelp(self, btn):
        """ Display a small help message box """
        helpMsg  = _('Any field of the form <b>{<tt>field</tt>}</b> will be replaced by the corresponding value. Available fields are:\n\n')
        helpMsg += track.getFormatSpecialFields()
        infoMsgBox(self.window, helpMsg)


    def onOk(self, btn):
        """ Save the new preferences """
        if not isdir(dirname(self.txtFile.get_text())):
            errorMsgBox(self.window, _('Invalid path'), _('The path to the selected file is not valid. Please choose an existing path.'))
            self.txtFile.grab_focus()
        else:
            prefs.set(__name__, 'file', self.txtFile.get_text())
            (start, end) = self.txtStatus.get_buffer().get_bounds()
            prefs.set(__name__, 'status', self.txtStatus.get_buffer().get_text(start, end))
            self.window.hide()


    def onOpen(self, btn):
        """ Let the user choose a file """
        dir    = dirname(self.txtFile.get_text())
        file   = basename(self.txtFile.get_text())
        result = fileChooser.save(self.window, _('Choose a file'), file, dir)

        if result is not None:
            self.txtFile.set_text(result)


    # --== Message handler ==--


    def handleMsg(self, msg, params):
        """ Handle messages sent to this module """
        if msg == modules.MSG_EVT_NEW_TRACK:
            self.updateFile(params['track'])
        elif msg == modules.MSG_EVT_STOPPED or msg == modules.MSG_EVT_APP_QUIT or msg == modules.MSG_EVT_MOD_UNLOADED:
            self.updateFile(None)
